REF_VALUE = '__REF_VALUE__'
FLAG_OF_REF = '__IS_REF__'
FLAG_OF_COMPUTED_REF = '__IS_COMPUTED_REF__'

REACTIVITY_VALUE = '__REACTIVITY_VALUE__'
FLAG_OF_REACTIVE = '__IS_REACTIVE__'

FLAG_OF_SKIP = '__REACTIVE_SKIP__'

FLAG_OF_READONLY = '__IS_READONLY__'
